/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Jean Le Feuvre 2000-2003 
 *					All rights reserved
 *
 *  This file is part of GPAC / Stream Management sub-project
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "SysCodecs.h"


static M4Err DUMMY_GetCapabilities(BaseDecoder *plug, CapObject *cap)
{
	cap->cap.valueINT = 0;
	return M4NotSupported;
}

static M4Err DUMMY_SetCapabilities(BaseDecoder *plug, const CapObject capability)
{
	return M4OK;
}




M4Err BIFS_AttachScene(SceneDecoder *plug, InlineScene *scene, Bool is_scene_decoder)
{
	BIFSPriv *priv = plug->privateStack;
	if (priv->codec) return M4BadParam;
	priv->pScene = scene;
	priv->app = scene->root_od->term;
	
	priv->codec = BIFS_NewDecoder(scene->graph, 0);
	/*attach the clock*/
	BIFS_SetClock(priv->codec, IS_GetSceneTime, scene);
	/*ignore all size info on anim streams*/
	if (!is_scene_decoder) BIFS_IgnoreSizeInfo(priv->codec);
	return M4OK;
}

static M4Err BIFS_AttachStream(BaseDecoder *plug, 
									 u16 ES_ID, 
									 unsigned char *decSpecInfo, 
									 u32 decSpecInfoSize, 
									 u16 DependsOnES_ID,
									 u32 objectTypeIndication, 
									 Bool Upstream)
{
	BIFSPriv *priv = plug->privateStack;
	if (Upstream) return M4NotSupported;
	return BIFS_ConfigureStream(priv->codec, ES_ID, decSpecInfo, decSpecInfoSize, objectTypeIndication);
}

static M4Err BIFS_DetachStream(BaseDecoder *plug, u16 ES_ID)
{
	BIFSPriv *priv = plug->privateStack;
	return BIFS_RemoveStream(priv->codec, ES_ID);
}

static M4Err BIFS_ProcessData(SceneDecoder*plug, unsigned char *inBuffer, u32 inBufferLength, 
								u16 ES_ID, u32 AU_time, u32 mmlevel)
{
	M4Err e = M4OK;
	BIFSPriv *priv = plug->privateStack;

	e = BIFS_DecodeAU(priv->codec, ES_ID, inBuffer, inBufferLength);

	/*if scene not attached do it*/
	IS_AttachGraphToRender(priv->pScene);
	return e;
}

void BIFSDec_Delete(BaseDecoder *plug)
{
	BIFSPriv *priv = plug->privateStack;
	BIFS_DeleteDecoder(priv->codec);
	free(priv);
	free(plug);
}

BaseDecoder *NewBIFSDecoder(u32 PL)
{
	BIFSPriv *priv;
	SceneDecoder *tmp;
	
	SAFEALLOC(tmp, sizeof(SceneDecoder));
	if (!tmp) return NULL;
	SAFEALLOC(priv, sizeof(BIFSPriv));
	priv->codec = NULL;
	priv->PL = PL;
	tmp->privateStack = priv;
	tmp->AttachStream = BIFS_AttachStream;
	tmp->DetachStream = BIFS_DetachStream;
	tmp->GetCapabilities = DUMMY_GetCapabilities;
	tmp->SetCapabilities = DUMMY_SetCapabilities;
	tmp->ProcessData = BIFS_ProcessData;
	tmp->AttachScene = BIFS_AttachScene;
	tmp->ReleaseScene = NULL;
	M4_REG_PLUG(tmp, M4SCENEDECODERINTERFACE, "GPAC BIFS Decoder", "gpac distribution", 0)
	return (BaseDecoder *) tmp;
}

typedef struct 
{
	u16 *ESIDs;
	u32 streamCount;
	LPODCODEC codec;
	InlineScene *scene;
	u32 PL;
} M4ODPriv;


static M4Err OD_AttachScene(SceneDecoder *plug, InlineScene *scene, Bool is_inline_scene)
{
	M4ODPriv *priv = plug->privateStack;
	if (priv->scene) return M4BadParam;
	priv->scene = scene;
	return M4OK;
}

static void ODS_SetupOD(InlineScene *is, ObjectDescriptor *od)
{
	ODManager *odm, *parent;
	odm = IS_FindODM(is, od->objectDescriptorID);
	/*remove the old OD*/
	if (odm) ODM_RemoveOD(odm);
	odm = NewODManager();
	odm->OD = od;
	odm->term = is->root_od->term;
	odm->parentscene = is;
	ChainAddEntry(is->ODlist, odm);

	/*locate service owner*/
	parent = is->root_od;
	while (parent->remote_OD) parent = parent->remote_OD;
	ODM_SetupOD(odm, parent->net_service);
}

static M4Err ODS_ODUpdate(M4ODPriv *priv, ObjectDescriptorUpdate *odU)
{
	ObjectDescriptor *od;
	u32 count;

	/*extract all our ODs and compare with what we already have...*/
	count = ChainGetCount(odU->objectDescriptors);
	if (count > 255) return M4InvalidDescriptor;

	while (count) {
		od = ChainGetEntry(odU->objectDescriptors, 0);
		ChainDeleteEntry(odU->objectDescriptors, 0);
		count--;
		ODS_SetupOD(priv->scene, od);
	}
	return M4OK;
}



static M4Err ODS_RemoveOD(M4ODPriv *priv, ObjectDescriptorRemove *odR)
{
	u32 i;
	for (i=0; i< odR->NbODs; i++) {
		ODManager *odm = IS_FindODM(priv->scene, odR->OD_ID[i]);
		if (odm) ODM_RemoveOD(odm);
	}
	return M4OK;
}

static M4Err ODS_UpdateESD(M4ODPriv *priv, ESDescriptorUpdate *ESDs)
{
	ESDescriptor *esd, *prev;
	ODManager *odm;
	u32 count, i;

	odm = IS_FindODM(priv->scene, ESDs->ODID);
	/*spec: "ignore"*/
	if (!odm) return M4OK;
	/*spec: "ES_DescriptorUpdate shall not be applied on object descriptors that have set URL_Flag to '1' (see 8.6.3)."*/
	if (odm->remote_OD) return M4NonCompliantBitStream;

	count = ChainGetCount(ESDs->ESDescriptors);

	while (count) {
		esd = ChainGetEntry(ESDs->ESDescriptors, 0);
		/*spec: "ES_Descriptors with ES_IDs that have already been received within the same name scope shall be ignored."*/
		prev = NULL;
		for (i=0; i<ChainGetCount(odm->OD->ESDescriptors); i++) {
			prev = ChainGetEntry(odm->OD->ESDescriptors, i);
			if (prev->ESID == esd->ESID) break;
			prev = NULL;
		}

		if (prev) {
			OD_DeleteDescriptor((Descriptor **)&esd);
		} else {
			/*and register new stream*/
			ChainAddEntry(odm->OD->ESDescriptors, esd);
			ODM_SetupStream(odm, esd, odm->net_service);
		}

		/*remove the desc from the AU*/
		ChainDeleteEntry(ESDs->ESDescriptors, 0);
		count--;
	}
	return M4OK;
}


static M4Err ODS_RemoveESD(M4ODPriv *priv, ESDescriptorRemove *ESDs)
{
	ODManager *odm;
	u32 i;
	odm = IS_FindODM(priv->scene, ESDs->ODID);
	/*spec: "ignore"*/
	if (!odm) return M4OK;
	/*spec: "ES_DescriptorRemove shall not be applied on object descriptors that have set URL_Flag to '1' (see 8.6.3)."*/
	if (odm->remote_OD) return M4NonCompliantBitStream;

	for (i=0; i<ESDs->NbESDs; i++) {
		/*blindly call remove*/
		ODM_RemoveStream(odm, ESDs->ES_ID[i]);
	}
	return M4OK;
}

static M4Err OD_AttachStream(BaseDecoder *plug, 
									 u16 ES_ID, 
									 unsigned char *decSpecInfo, 
									 u32 decSpecInfoSize, 
									 u16 DependsOnES_ID,
									 u32 objectTypeIndication, 
									 Bool Upstream)
{
	M4ODPriv *priv = plug->privateStack;
	
	if (Upstream) return M4OK;
	//warning, we only support one stream ...
	if (priv->streamCount) return M4NotSupported;
	priv->ESIDs = malloc(sizeof(u16));
	priv->ESIDs[0] = ES_ID;
	priv->streamCount ++;
	//that's it, OD doesn't use decoderSpecificInfo
	return M4OK;
}


static M4Err OD_DetachStream(BaseDecoder *plug, u16 ES_ID)
{
	M4ODPriv *priv = plug->privateStack;

	//warning, we only support one stream ...
	if (!priv->streamCount) return M4BadParam;
	free(priv->ESIDs);
	priv->ESIDs = NULL;
	priv->streamCount = 0;

	return M4OK;
}


static M4Err OD_ProcessData(SceneDecoder *plug, unsigned char *inBuffer, u32 inBufferLength, 
							u16 ES_ID, u32 AU_time, u32 mmlevel)
{
	M4Err e;
	ODCommand *com;
	M4ODPriv *priv = plug->privateStack;
	
	//OD codec directly manipulates the terminal resources. No need for
	//size request

	//1- set up the decoder
	e = OD_SetBuffer(priv->codec, inBuffer, inBufferLength);
	if (e) return e;
	//2- decode the AU
	e = OD_DecodeAU(priv->codec);
	if (e) return e;

	//3- process all the commands in this AU, in order
	while (1) {
		com = OD_GetCommand(priv->codec);
		if (!com) break;

		//ok, we have a command
		switch (com->tag) {
		case ODUpdate_Tag:
			e = ODS_ODUpdate(priv, (ObjectDescriptorUpdate *) com);
			break;
		case ODRemove_Tag:
			e = ODS_RemoveOD(priv, (ObjectDescriptorRemove *) com);
			break;
		case ESDUpdate_Tag:
			e = ODS_UpdateESD(priv, (ESDescriptorUpdate *)com);
			break;
		case ESDRemove_Tag:
			e = ODS_RemoveESD(priv, (ESDescriptorRemove *)com);
			break;
		case IPMPDUpdate_Tag:
			e = M4NotSupported;
			break;
		case IPMPDRemove_Tag:
			e = M4NotSupported;
			break;
		/*should NEVER exist outside the file format*/
		case ESDRemove_Ref_Tag:
			e = M4NonCompliantBitStream;
			break;
		default:
			if (com->tag >= ISO_RESERVED_COMMANDS_BEGIN && com->tag <= ISO_RESERVED_COMMANDS_END) {
				e = M4ISOForbiddenDescriptor;
			} else {
				/*we don't process user commands*/
				e = M4OK;
			}
			break;
		}
		OD_DeleteCommand(&com);
		if (e) return e;
	}
	return M4OK;
}

void ODDec_Delete(BaseDecoder *plug)
{
	M4ODPriv *priv = plug->privateStack;

	if (priv->ESIDs) free(priv->ESIDs);
	OD_DeleteCodec(priv->codec);
	free(priv);
	free(plug);
}

BaseDecoder *NewODDecoder(u32 PL)
{
	SceneDecoder *tmp;
	M4ODPriv *priv;
	
	SAFEALLOC(tmp, sizeof(SceneDecoder));
	if (!tmp) return NULL;
	SAFEALLOC(priv, sizeof(M4ODPriv));
	priv->ESIDs = NULL;
	priv->streamCount = 0;
	//we only work with decoding ...
	priv->codec = OD_NewCodec(OD_READ);
	priv->scene = NULL;
	priv->PL = PL;

	tmp->privateStack = priv;
	tmp->AttachStream = OD_AttachStream;
	tmp->DetachStream = OD_DetachStream;
	tmp->GetCapabilities = DUMMY_GetCapabilities;
	tmp->SetCapabilities = DUMMY_SetCapabilities;
	tmp->ProcessData = OD_ProcessData;
	tmp->AttachScene = OD_AttachScene;

	M4_REG_PLUG(tmp, M4SCENEDECODERINTERFACE, "GPAC OD Decoder", "gpac distribution", 0)
	return (BaseDecoder *) tmp;
}




#include <gpac/m4_author.h>


typedef struct 
{
	InlineScene *inline_scene;
	M4Client *app;
	M4SceneManager *ctx;
	char *fileName;
	u32 load_flags;
	u32 nb_streams;
	u32 base_stream_id;
	/*mp3 import from flash*/
	Chain *files_to_delete;
	M4ContextLoader load;
	u32 type;
} M4CTXPriv;

static M4Err M4CTX_GetCapabilities(BaseDecoder *plug, CapObject *cap)
{
	return M4NotSupported;
}


static void M4CTX_Reset(M4CTXPriv *priv)
{
	if (priv->ctx) M4SM_Delete(priv->ctx);
	priv->ctx = NULL;
	SG_Reset(priv->inline_scene->graph);
	if (priv->load_flags != 3) priv->load_flags = 0;
	while (ChainGetCount(priv->files_to_delete)) {
		char *fileName = ChainGetEntry(priv->files_to_delete, 0);
		ChainDeleteEntry(priv->files_to_delete, 0);
		M4_DeleteFile(fileName);
		free(fileName);
	}
}

void M4CTX_OnActivate(SFNode *node)
{
	M4CTXPriv *priv = (M4CTXPriv *) Node_GetPrivate(node);
	B_Conditional*c = (B_Conditional*)node;
	/*always apply in parent graph to handle protos correctly*/
	if (c->activate) SG_ApplyCommandList(Node_GetParentGraph(node), c->buffer.commandList, IS_GetSceneTime(priv->inline_scene));
}
void M4CTX_OnReverseActivate(SFNode *node)
{
	M4CTXPriv *priv = (M4CTXPriv *) Node_GetPrivate(node);
	B_Conditional*c = (B_Conditional*)node;
	/*always apply in parent graph to handle protos correctly*/
	if (!c->reverseActivate) SG_ApplyCommandList(Node_GetParentGraph(node), c->buffer.commandList, IS_GetSceneTime(priv->inline_scene));
}

void M4CTX_NodeInit(void *cbk, SFNode *node)
{
	M4CTXPriv *priv = (M4CTXPriv *) cbk;
	if (Node_GetTag(node) == TAG_Conditional) {
		((B_Conditional*)node)->on_activate = M4CTX_OnActivate;
		((B_Conditional*)node)->on_reverseActivate = M4CTX_OnReverseActivate;
		Node_SetPrivate(node, priv);
	} else {
		Term_NodeInit(priv->inline_scene, node);
	}
}

static M4Err M4CTX_AttachScene(SceneDecoder *plug, InlineScene *scene, Bool is_scene_decoder)
{
	M4CTXPriv *priv = plug->privateStack;
	if (priv->ctx) return M4BadParam;

	priv->inline_scene = scene;
	priv->app = scene->root_od->term;

	SG_SetInitCallback(scene->graph, M4CTX_NodeInit, priv);
	return M4OK;
}

static M4Err M4CTX_ReleaseScene(SceneDecoder *plug)
{
	M4CTX_Reset((M4CTXPriv *) plug->privateStack);
	return M4OK;
}

static M4Err M4CTX_AttachStream(BaseDecoder *plug, 
									 u16 ES_ID, 
									 unsigned char *decSpecInfo, 
									 u32 decSpecInfoSize, 
									 u16 DependsOnES_ID,
									 u32 objectTypeIndication, 
									 Bool Upstream)
{
	M4CTXPriv *priv = plug->privateStack;
	if (Upstream) return M4NotSupported;

	/*animation stream like*/
	if (priv->ctx) {
		u32 i;
		for (i=0; i<ChainGetCount(priv->ctx->streams); i++) {
			M4StreamContext *sc = ChainGetEntry(priv->ctx->streams, i);
			if (ES_ID == sc->ESID) {
				priv->nb_streams++;
				return M4OK;
			}
		}
		return M4NonCompliantBitStream;
	}
	/*main dummy stream we need a dsi*/
	if (!decSpecInfo) return M4NonCompliantBitStream;

	priv->fileName = strdup(decSpecInfo);
	priv->nb_streams = 1;
	priv->load_flags = 0;
	priv->base_stream_id = ES_ID;
	return M4OK;
}

static M4Err M4CTX_DetachStream(BaseDecoder *plug, u16 ES_ID)
{
	M4CTXPriv *priv = plug->privateStack;
	priv->nb_streams --;
	return M4OK;
}

static Bool M4CTX_StreamInRootOD(ObjectDescriptor *od, u32 ESID)
{
	u32 i, count;
	/*no root, only one stream possible*/
	if (!od) return 1;
	count = ChainGetCount(od->ESDescriptors);
	/*idem*/
	if (!count) return 1;
	for (i=0; i<count; i++) {
		ESDescriptor *esd = ChainGetEntry(od->ESDescriptors, i);
		if (esd->ESID==ESID) return 1;
	}
	return 0;
}

static void M4CTX_OnMessage(void *cbk, char *szMsg, M4Err e)
{
	M4CTXPriv *priv = (M4CTXPriv *)cbk;
	if (e) M4_OnMessage(priv->inline_scene->root_od->term, priv->inline_scene->root_od->net_service->url, szMsg, e);
}

static void M4CTX_OnProgress(void *cbk, u32 done, u32 tot)
{
	char szMsg[1024];
	M4CTXPriv *priv = (M4CTXPriv *)cbk;
	sprintf(szMsg, "Importing %d / %d (%.2f %%)", done, tot, ((Float)100*done)/tot);
	M4_OnMessage(priv->inline_scene->root_od->term, 
		priv->inline_scene->root_od->net_service->url, szMsg, M4OK);
}

static M4Err M4CTX_ProcessData(SceneDecoder *plug, unsigned char *inBuffer, u32 inBufferLength, 
								u16 ES_ID, u32 stream_time, u32 mmlevel)
{
	M4Err e = M4OK;
	u32 i, j, k, nb_updates, max_dur;
	M4AUContext *au;
	Bool can_delete_com;
	M4StreamContext *sc;
	M4CTXPriv *priv = plug->privateStack;

	/*something failed*/
	if (priv->load_flags==3) return M4EOF;

	/*this signals main scene deconnection, destroy the context if needed*/
	assert(ES_ID);

	if (priv->load_flags != 2) {
		/*load first frame only*/
		if (!priv->load_flags) {
			priv->load_flags = 1;
			priv->ctx = NewSceneManager(priv->inline_scene->graph);
			memset(&priv->load, 0, sizeof(M4ContextLoader));
			priv->load.ctx = priv->ctx;
			priv->load.cbk = priv;
			priv->load.scene_graph = priv->inline_scene->graph;
			priv->load.fileName = priv->fileName;
			priv->load.OnMessage = M4CTX_OnMessage;
			priv->load.OnProgress = M4CTX_OnProgress;
			priv->load.load_scripts = 1;
			priv->load.import_flags = M4SWF_StaticDictionary | M4SWF_UseXCurve | M4SWF_UseXLineProps;
			e = M4SM_LoaderInit(&priv->load);
			if (!e) 
				SG_SetSizeInfo(priv->inline_scene->graph, priv->ctx->scene_width, priv->ctx->scene_height, priv->ctx->is_pixel_metrics);
		} 
		/*load the rest*/
		else {
			priv->load_flags = 2;
			e = M4SM_LoaderRun(&priv->load);
			M4SM_LoaderDone(&priv->load);
		}

		if (e) {
			M4SM_Delete(priv->ctx);
			priv->ctx = NULL;
			priv->load_flags = 3;
			return e;
		}

		/*and figure out duration of root scene, and take care of XMT timing*/
		max_dur = 0;
		for (i=0; i<ChainGetCount(priv->ctx->streams); i++) {
			sc = ChainGetEntry(priv->ctx->streams, i);
			/*all streams in root OD are handled with ESID 0 to differentiate with any animation streams*/
			if (M4CTX_StreamInRootOD(priv->ctx->root_od, sc->ESID)) sc->ESID = 0;
			if (!sc->timeScale) sc->timeScale = 1000;

			au = NULL;
			for (j = 0; j<ChainGetCount(sc->AUs); j++) {
				au = ChainGetEntry(sc->AUs, j);
				if (!au->timing) au->timing = (u32) (sc->timeScale*au->timing_sec);
			}
			if (au && !sc->ESID && (au->timing>max_dur)) max_dur = au->timing * 1000 / sc->timeScale;
		}
		if (priv->load_flags==2) {
			if (max_dur) {
				priv->inline_scene->root_od->duration = max_dur;
				IS_SetSceneDuration(priv->inline_scene);
			}
		}
	}

	nb_updates = 0;

	for (i=0; i<ChainGetCount(priv->ctx->streams); i++) {
		M4StreamContext *sc = ChainGetEntry(priv->ctx->streams, i);
		/*not our stream*/
		if (sc->ESID && (sc->ESID != ES_ID)) continue;
		/*not the base stream*/
		if (!sc->ESID && (priv->base_stream_id != ES_ID)) continue;
		/*handle SWF media extraction*/
		if ((sc->streamType == M4ST_OD) && (priv->load_flags==1)) continue;

		/*check for seek*/
		if (sc->last_au_time > 1 + stream_time) {
			/*root streams shall not seek this way*/
			if (!sc->ESID) assert(0);
			/*on other streams, we assume the stream is 100% RAP wrt to the base stream*/
			else {
				sc->last_au_time = 0;
			}
		}

		can_delete_com = 0;
		if (!sc->ESID && (priv->load_flags==2)) can_delete_com = 1;

		/*we're in the right stream, apply update*/
		for (j=0; j<ChainGetCount(sc->AUs); j++) {
			M4AUContext *au = ChainGetEntry(sc->AUs, j);
			u32 au_time = au->timing*1000/sc->timeScale;
			if (au_time + 1 <= sc->last_au_time) {
				/*remove first replace command*/
				if (!sc->ESID && (sc->streamType==M4ST_SCENE)) {
					while (ChainGetCount(au->commands)) {
						SGCommand *com = ChainGetEntry(au->commands, 0);
						ChainDeleteEntry(au->commands, 0);
						SG_DeleteCommand(com);
					}
					ChainDeleteEntry(sc->AUs, j);
					DeleteChain(au->commands);
					free(au);
					j--;
				}
				continue;
			}
			if (au_time > stream_time) {
				nb_updates++;
				break;
			}

			if (sc->streamType == M4ST_SCENE) {
				/*apply the commands*/
				for (k=0; k<ChainGetCount(au->commands); k++) {
					SGCommand *com = ChainGetEntry(au->commands, k);
					e = SG_ApplyCommand(priv->inline_scene->graph, com, 0);
					if (e) break;
					/*remove commands on base layer*/
					if (can_delete_com) {
						ChainDeleteEntry(au->commands, k);
						SG_DeleteCommand(com);
						k--;
					}
				}
			} 
			else if (sc->streamType == M4ST_OD) {
				/*apply the commands*/
				while (ChainGetCount(au->commands)) {
					Bool keep_com = 0;
					ODCommand *com = ChainGetEntry(au->commands, 0);
					ChainDeleteEntry(au->commands, 0);
					switch (com->tag) {
					case ODUpdate_Tag:
					{
						ObjectDescriptorUpdate *odU = (ObjectDescriptorUpdate *)com;
						while (ChainGetCount(odU->objectDescriptors)) {
							ESDescriptor *esd;
							char *remote;
							MuxInfoDescriptor *mux = NULL;
							ObjectDescriptor *od = ChainGetEntry(odU->objectDescriptors, 0);
							ChainDeleteEntry(odU->objectDescriptors, 0);
							/*we can only work with single-stream ods*/
							esd = ChainGetEntry(od->ESDescriptors, 0);
							if (!esd) {
								if (od->URLString) {
									ODS_SetupOD(priv->inline_scene, od);
								} else {
									OD_DeleteDescriptor((Descriptor **) &od);
								}
								continue;
							}
							/*fix OCR dependencies*/
							if (M4CTX_StreamInRootOD(priv->ctx->root_od, esd->OCRESID)) esd->OCRESID = priv->base_stream_id;

							/*forbidden if ESD*/
							if (od->URLString) {
								OD_DeleteDescriptor((Descriptor **) &od);
								continue;
							}
							/*look for MUX info*/
							for (k=0; k<ChainGetCount(esd->extensionDescriptors); k++) {
								mux = ChainGetEntry(esd->extensionDescriptors, k);
								if (mux->tag == MuxInfoDescriptor_Tag) break;
								mux = NULL;
							}
							/*we need a mux if not animation stream*/
							if (!mux || !mux->file_name) {
								/*only animation streams are handled*/
								if (!esd->decoderConfig) {
									OD_DeleteDescriptor((Descriptor **) &od);
								} else if (esd->decoderConfig->streamType==M4ST_SCENE) {
									/*set ST to private scene to get sure the stream will be redirected to us*/
									esd->decoderConfig->streamType = M4ST_PRIVATE_SCENE;
									esd->dependsOnESID = priv->base_stream_id;
									ODS_SetupOD(priv->inline_scene, od);
								} else if (esd->decoderConfig->streamType==M4ST_INTERACT) {
									UIConfigDescriptor *cfg = (UIConfigDescriptor *) esd->decoderConfig->decoderSpecificInfo;
									OD_EncodeUIConfig(cfg, &esd->decoderConfig->decoderSpecificInfo);
									OD_DeleteDescriptor((Descriptor **) &cfg);
									ODS_SetupOD(priv->inline_scene, od);
								} else {
									OD_DeleteDescriptor((Descriptor **) &od);
								}
								continue;
							}
							/*text import*/
							if (mux->textNode) {
								M4SM_ImportSRT_BIFS(priv->ctx, esd, mux);
								/*set ST to private scene and dependency to base to get sure the stream will be redirected to us*/
								esd->decoderConfig->streamType = M4ST_PRIVATE_SCENE;
								esd->dependsOnESID = priv->base_stream_id;
								ODS_SetupOD(priv->inline_scene, od);
								continue;
							}

							/*soundstreams are a bit of a pain, they may be declared before any data gets written*/
							if (mux->delete_file) {
								FILE *t = fopen(mux->file_name, "rb");
								if (!t) {
									keep_com = 1;
									ChainInsertEntry(odU->objectDescriptors, od, 0);
									break;
								}
								fclose(t);
							}
							/*remap to remote URL*/
							remote = strdup(mux->file_name);
							k = od->objectDescriptorID;
							/*if files were created we'll have to clean up (swf import)*/
							if (mux->delete_file) ChainAddEntry(priv->files_to_delete, strdup(remote));

							OD_DeleteDescriptor((Descriptor **) &od);
							od = (ObjectDescriptor *) OD_NewDescriptor(ObjectDescriptor_Tag);
							od->URLString = remote;
							od->objectDescriptorID = k;
							ODS_SetupOD(priv->inline_scene, od);
						}
						if (keep_com) break;
					}
						break;
					case ODRemove_Tag:
					{
						ObjectDescriptorRemove *odR = (ObjectDescriptorRemove*)com;
						for (k=0; k<odR->NbODs; k++) {
							ODManager *odm = IS_FindODM(priv->inline_scene, odR->OD_ID[k]);
							if (odm) ODM_RemoveOD(odm);
						}
					}
						break;
					default:
						break;
					}
					if (keep_com) {
						ChainInsertEntry(au->commands, com, 0);
						break;
					} else {
						OD_DeleteCommand(&com);
					}
					if (e) break;
				}

			}
			sc->last_au_time = au_time + 1;
			/*attach graph to renderer*/
			IS_AttachGraphToRender(priv->inline_scene);
			if (e) return e;

			/*for root streams remove completed AUs (no longer needed)*/
			if (!sc->ESID && !ChainGetCount(au->commands) ) {
				ChainDeleteEntry(sc->AUs, j);
				DeleteChain(au->commands);
				free(au);
				j--;
			}
		}
	}
	if (e) return e;
	if ((priv->load_flags==2) && !nb_updates) return M4EOF;
	return M4OK;
}

const char *M4CTX_GetName(struct _basedecoder *plug)
{
	M4CTXPriv *priv = plug->privateStack;

	switch (priv->load.loader_type) {
	case 1: return "GPAC BT/VRML Parser";
	case 2: return "GPAC XMT Parser";
	case 3: return "GPAC MP4 Memory Loader";
	case 4: return "GPAC Flash/SWF Emulator";
	case 5: return "GPAC SVG Parser";
	}
	return "Unknown";
}

Bool M4CTX_CanHandleStream(BaseDecoder *ifce, u32 StreamType, u32 ObjectType, unsigned char *decSpecInfo, u32 decSpecInfoSize, u32 PL)
{
	if ((StreamType==M4ST_PRIVATE_SCENE) && (ObjectType==1)) return 1;
	return 0;
}

void DeleteContextLoader(BaseDecoder *plug)
{
	M4CTXPriv *priv = plug->privateStack;
	if (priv->fileName) free(priv->fileName);
	assert(!priv->ctx);
	DeleteChain(priv->files_to_delete);
	free(priv);
	free(plug);
}

BaseDecoder *NewContextLoader()
{
	M4CTXPriv *priv;
	SceneDecoder *tmp;
	
	SAFEALLOC(tmp, sizeof(SceneDecoder));
	SAFEALLOC(priv, sizeof(M4CTXPriv));
	priv->files_to_delete = NewChain();

	tmp->privateStack = priv;
	tmp->AttachStream = M4CTX_AttachStream;
	tmp->DetachStream = M4CTX_DetachStream;
	tmp->GetCapabilities = M4CTX_GetCapabilities;
	tmp->SetCapabilities = DUMMY_SetCapabilities;
	tmp->ProcessData = M4CTX_ProcessData;
	tmp->AttachScene = M4CTX_AttachScene;
	tmp->ReleaseScene = M4CTX_ReleaseScene;
	tmp->GetName = M4CTX_GetName;
	tmp->CanHandleStream = M4CTX_CanHandleStream;
	M4_REG_PLUG(tmp, M4SCENEDECODERINTERFACE, "GPAC Context Loader", "gpac distribution", 0)
	return (BaseDecoder*)tmp;
}
