/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Jean Le Feuvre 2000-2003 
 *					All rights reserved
 *
 *  This file is part of GPAC / 2D rendering plugin
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "generic_text.h"

void text2D_get_ascent_descent(DrawableContext *ctx, Float *a, Float *d)
{
	TextStack2D *stack = (TextStack2D *) Node_GetPrivate(ctx->node->owner);
	*a = stack->ascent;
	*d = stack->descent;
}


TextLineEntry2D *NewTextLine2D(Render2D *sr)
{
	TextLineEntry2D *tl;
	SAFEALLOC(tl, sizeof(TextLineEntry2D));
	tl->path = m4_new_path();
	tl->path->fineness = 1.0;
	/*text texturing enabled*/
	if (sr->compositor->use_texture_text) {
		tl->sr = sr;
		tl->hwtx = sr->compositor->g_hw->new_stencil(sr->compositor->g_hw, M4StencilTexture);
		tl->last_scale.x = sr->scale_x;
		tl->last_scale.y = sr->scale_y;
	}
	return tl;
}

Bool TextLine2D_TextureIsReady(TextLineEntry2D *tl)
{
	M4Matrix2D mx;
	LPM4STENCIL stenc;
	LPM4SURFACE surf;
	Float cx, cy;
	u32 width, height;
	Float scale;
	M4Err e;
	Graphics2DDriver *g_hw = tl->sr->compositor->g_hw;
	
	/*something failed*/
	if (!tl->hwtx) return 0;

	if (tl->tx_ready) {
		if ((tl->last_scale.x == tl->sr->scale_x)
			&& (tl->last_scale.y == tl->sr->scale_y)
			) return 1;

		if (tl->hwtx) g_hw->delete_stencil(tl->hwtx);
		if (tl->tx_path) m4_path_delete(tl->tx_path);
		tl->tx_path = NULL;
		tl->hwtx = g_hw->new_stencil(g_hw, M4StencilTexture);

		tl->last_scale.x = tl->sr->scale_x;
		tl->last_scale.y = tl->sr->scale_y;
	}

	scale = MAX(tl->last_scale.x, tl->last_scale.y);
	if ((scale*tl->bounds.width>512) || (scale*tl->bounds.height>512)) {
		scale = MIN(512/tl->bounds.width, 512/tl->bounds.height);
	}

	tl->bounds.x = tl->path->min_x;
	tl->bounds.y = tl->path->max_y;
	tl->bounds.width = (tl->path->max_x-tl->path->min_x);
	tl->bounds.height = (tl->path->max_y-tl->path->min_y);


	width = (u32) ceil(scale*tl->bounds.width);
	height = (u32) ceil(scale*tl->bounds.height);
	width += 4;
	height += 4;

	surf = g_hw->new_surface(g_hw, 1);
	if (!surf) {
		g_hw->delete_stencil(tl->hwtx);
		tl->hwtx = NULL;
		return 0;
	}
	/*FIXME - make it work with alphagrey...*/
	e = g_hw->stencil_create_texture(tl->hwtx, width, height, M4PF_ARGB);
	if (!e) e = g_hw->attach_surface_to_texture(surf, tl->hwtx);
	g_hw->surface_clear(surf, NULL, 0);
	stenc = g_hw->new_stencil(g_hw, M4StencilSolid);
	g_hw->set_brush_color(stenc, 0xFF000000);

	cx = (tl->path->max_x + tl->path->min_x) /2;
	cy = (tl->path->max_y + tl->path->min_y) /2;

	mx2d_init(mx);
	mx2d_add_translation(&mx, -cx, -cy);
	mx2d_add_scale(&mx, scale, scale);
	g_hw->surface_set_matrix(surf, &mx);

	g_hw->surface_set_raster_level(surf, M4RasterHighQuality);
	g_hw->surface_set_path(surf, tl->path);
	g_hw->surface_fill(surf, stenc);
	g_hw->delete_stencil(stenc);
	g_hw->delete_surface(surf);

	tl->tx_path = m4_new_path();
	m4_path_add_move_to(tl->tx_path, tl->path->min_x, tl->path->min_y);
	m4_path_add_line_to(tl->tx_path, tl->path->max_x, tl->path->min_y);
	m4_path_add_line_to(tl->tx_path, tl->path->max_x, tl->path->max_y);
	m4_path_add_line_to(tl->tx_path, tl->path->min_x, tl->path->max_y);
	m4_path_close(tl->tx_path);

	tl->tx_bounds.x = tl->tx_bounds.y = 0;
	tl->tx_bounds.width = (Float) width;
	tl->tx_bounds.height = (Float) height;

	if (e) {
		g_hw->delete_stencil(tl->hwtx);
		tl->hwtx = NULL;
		return 0;
	}
	tl->tx_ready = 1;
	return 1;
}

void TextStack2D_clean_paths(TextStack2D *stack)
{
	TextLineEntry2D *tl;
	/*delete all path objects*/
	while (ChainGetCount(stack->text_lines)) {
		tl = ChainGetEntry(stack->text_lines, 0);
		ChainDeleteEntry(stack->text_lines, 0);
		if (tl->path) m4_path_delete(tl->path);
		if (tl->hwtx) tl->sr->compositor->g_hw->delete_stencil(tl->hwtx);
		if (tl->tx_path) m4_path_delete(tl->tx_path);
		free(tl);
	}
	m4_rect_reset(&stack->bounds);
	drawable_reset_path(stack->graph);
}

void Text2D_Draw(DrawableContext *ctx)
{
	u32 i;
	Bool can_texture_text;
	TextLineEntry2D *tl;
	TextStack2D *st = (TextStack2D *) Node_GetPrivate((SFNode *) ctx->node->owner);

	/*text has been splited*/
	if (ctx->sub_path_index > 0) {
		tl = ChainGetEntry(st->text_lines, ctx->sub_path_index - 1);
		if (!tl || !tl->path) return;
		VS2D_TexturePath(ctx->surface, tl->path, ctx);
		VS2D_DrawPath(ctx->surface, tl->path, ctx, NULL, NULL);
		return;
	}

	can_texture_text = 0;
	if (st->graph->compositor->use_texture_text || (st->texture_text_flag==2)) {
		can_texture_text = !ctx->h_texture && !ctx->aspect.pen_props.width;
	}

	for (i=0; i<ChainGetCount(st->text_lines); i++) {
		tl = ChainGetEntry(st->text_lines, i);

		if (can_texture_text  && TextLine2D_TextureIsReady(tl)) {
			VS2D_TexturePathText(ctx->surface, ctx, tl->tx_path, &tl->bounds, tl->hwtx, &tl->tx_bounds);
		} else {
			VS2D_TexturePath(ctx->surface, tl->path, ctx);
			VS2D_DrawPath(ctx->surface, tl->path, ctx, NULL, NULL);
		}
		/*reset fill/strike flags since we perform several draw per context*/
		ctx->path_filled = ctx->path_stroke = 0;
	}
}

Bool Text2D_PointOver(DrawableContext *ctx, Float x, Float y, Bool check_outline)
{
	M4Matrix2D inv;
	u32 i;
	TextLineEntry2D *tl;
	TextStack2D *st;
	/*this is not documented anywhere but it speeds things up*/
	if (!check_outline) return 1;
	
	
	st = (TextStack2D *) Node_GetPrivate((SFNode *) ctx->node->owner);
	
	mx2d_copy(inv, ctx->transform);
	mx2d_inverse(&inv);
	mx2d_apply_coords(&inv, &x, &y);

	/*otherwise get all paths*/
	if (ctx->sub_path_index > 0) {
		tl = ChainGetEntry(st->text_lines, ctx->sub_path_index - 1);
		if (!tl || !tl->path) return 0;
		return m4_path_point_over(tl->path, x, y);
	}

	for (i=0; i<ChainGetCount(st->text_lines); i++) {
		tl = ChainGetEntry(st->text_lines, i);
		if (!tl->path) return 0;
		if (m4_path_point_over(tl->path, x, y)) return 1;
	}
	return 0;
}
