/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Jean Le Feuvre 2000-2004 
 *					All rights reserved
 *
 *  This file is part of GPAC / 3D rendering plugin
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "visual_surface.h"

#include "gl_inc.h"

#define CHECK_GL_EXT(name) ((strstr(ext, name) != NULL) ? 1 : 0)
void R3D_LoadExtensions(Render3D *sr)
{
	const char *ext = glGetString(GL_EXTENSIONS);
	if (!ext) return;
	memset(&sr->compositor->hw_caps, 0, sizeof(HardwareCaps));

	if (CHECK_GL_EXT("GL_ARB_multisample") || CHECK_GL_EXT("GLX_ARB_multisample") || CHECK_GL_EXT("WGL_ARB_multisample")) 
		sr->compositor->hw_caps.multisample = 1;
	if (CHECK_GL_EXT("GL_ARB_texture_non_power_of_two")) 
		sr->compositor->hw_caps.npot_texture = 1;
	if (CHECK_GL_EXT("GL_EXT_abgr")) 
		sr->compositor->hw_caps.abgr_texture = 1;
	if (CHECK_GL_EXT("GL_EXT_bgra")) 
		sr->compositor->hw_caps.bgra_texture = 1;

	if (CHECK_GL_EXT("GL_EXT_texture_rectangle") || CHECK_GL_EXT("GL_NV_texture_rectangle")) 
		sr->compositor->hw_caps.rect_texture = 1;
}


void VS3D_Setup(VisualSurface *surf)
{
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glClearDepth(1.0f);
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (float) (0.2 * 128));

    glShadeModel(GL_SMOOTH);
	glGetIntegerv(GL_MAX_LIGHTS, &surf->max_lights);

	if (surf->render->compositor->high_speed) {
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
		glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
		glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
		glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
	} else {
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
		glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
		glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	}

	if (surf->render->compositor->antiAlias == M4_AL_All) {
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_POINT_SMOOTH);
		if (surf->render->poly_aa)
			glEnable(GL_POLYGON_SMOOTH);
		else
			glDisable(GL_POLYGON_SMOOTH);
	} else {
		glDisable(GL_LINE_SMOOTH);
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_POLYGON_SMOOTH);
	}

	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glDisable(GL_FOG);
	glClear(GL_DEPTH_BUFFER_BIT);
}

void VS3D_SetHeadlight(VisualSurface *surf, Bool bOn, M4Camera *cam)
{
	SFVec3f dir;
	SFColor col;
	if (!bOn) return;

	col.blue = col.red = col.green = 1.0;
	if (cam->is_3D) {
		dir = camera_get_target_dir(cam);
	} else {
		dir.x = dir.y = 0; dir.z = 1;
	}
	VS3D_AddDirectionalLight(surf, 0.0, col, 0.9f, dir);
}

void VS3D_SetViewport(VisualSurface *surf, M4Rect vp)
{
	glViewport((s32) vp.x, (s32) vp.y, (s32) vp.width, (s32) vp.height);
}

void VS3D_ClearDepth(VisualSurface *surf)
{
	glClear(GL_DEPTH_BUFFER_BIT);
}

void VS3D_DrawAABBNode(RenderEffect3D *eff, M4Mesh *mesh, u32 prim_type, AABBNode *n)
{
	u32 i;
	u32 cull = camera_cull_box(eff->camera, &eff->model_matrix, n->min, n->max);
	if (cull==CULL_OUTSIDE) return;
	if ((cull==CULL_INTERSECTS) && n->pos) {
		VS3D_DrawAABBNode(eff, mesh, prim_type, n->pos);
		VS3D_DrawAABBNode(eff, mesh, prim_type, n->neg);
	}
	/*the good thing about the structure of the aabb tree is that the primitive idx is vali for both
	leaf and non-leaf nodes, so we can use it as soon as we have a CULL_INSIDE.
	However we must push triangles one by one since primitive order may have been swapped when
	building the AABB tree*/
	for (i=0; i<n->nb_idx; i++) {
		glDrawElements(prim_type, 3, GL_UNSIGNED_INT, &mesh->indices[3*n->indices[i]]);
	}
}

void VS3D_DrawMeshIntern(RenderEffect3D *eff, M4Mesh *mesh)
{
	u32 prim_type;
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, sizeof(M4Vertex),  &mesh->vertices[0].pos);

	if (mesh->flags & MESH_HAS_COLOR) {
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(3, GL_FLOAT, sizeof(M4Vertex), &mesh->vertices[0].color);
	} else {
		glDisableClientState(GL_COLOR_ARRAY);
	}

	if (!mesh->mesh_type && !(mesh->flags & MESH_NO_TEXTURE)) {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY );
		glTexCoordPointer(2, GL_FLOAT, sizeof(M4Vertex), &mesh->vertices[0].texcoords);
	}
	
	if (mesh->mesh_type || (mesh->flags & MESH_IS_2D)) {
		glDisableClientState(GL_NORMAL_ARRAY);
		if (mesh->mesh_type) glDisable(GL_LIGHTING);
		glNormal3f(0, 0, 1.0f);
		glDisable(GL_CULL_FACE);
	} else {
		glEnableClientState(GL_NORMAL_ARRAY );
		glNormalPointer(GL_FLOAT, sizeof(M4Vertex), &mesh->vertices[0].normal);

		if (!mesh->mesh_type) {
			/*if mesh is transparent DON'T CULL*/
			if (!eff->mesh_is_transparent && (mesh->flags & MESH_IS_SOLID)) {
				glEnable(GL_CULL_FACE);
				glFrontFace((mesh->flags & MESH_IS_CW) ? GL_CW : GL_CCW);
			} else {
				glDisable(GL_CULL_FACE);
			}
		}
	}

	switch (mesh->mesh_type) {
	case MESH_LINESET:
		prim_type = GL_LINES;
		break;
	case MESH_POINTSET:
		prim_type = GL_POINTS;
		break;
	default:
		prim_type = GL_TRIANGLES;
		break;
	}

	/*gain is not obvious on small models, and really worse with large ones - we'd need more than AABB culling*/
#if 0
	if ((eff->cull_flag==CULL_INSIDE) || !mesh->aabb_root || !mesh->aabb_root->pos) {
		glDrawElements(prim_type, mesh->i_count, GL_UNSIGNED_INT, mesh->indices);
	} else {
		VS3D_DrawAABBNode(eff, mesh, prim_type, mesh->aabb_root->pos);
		VS3D_DrawAABBNode(eff, mesh, prim_type, mesh->aabb_root->neg);
	}
#else
	glDrawElements(prim_type, mesh->i_count, GL_UNSIGNED_INT, mesh->indices);
#endif

	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY );
	glDisableClientState(GL_NORMAL_ARRAY );

	eff->mesh_is_transparent = 0;
}

void VS3D_DrawAABBNodeBounds(RenderEffect3D *eff, AABBNode *node)
{
	if (node->pos) {
		VS3D_DrawAABBNodeBounds(eff, node->pos);
		VS3D_DrawAABBNodeBounds(eff, node->neg);
	} else {
		SFVec3f c, s;
		vec_diff(s, node->max, node->min);
		vec_scale(c, s, 0.5f);
		vec_add(c, node->min, c);

		glPushMatrix();
		glTranslatef(c.x, c.y, c.z);
		glScalef(s.x, s.y, s.z);
		VS3D_DrawMeshIntern(eff, eff->surface->render->unit_bbox);
		glPopMatrix();
	}
}

void VS3D_DrawMeshAABBTreeBounds(RenderEffect3D *eff, M4Mesh *mesh)
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glColor3f(0.0, 0.0, 0.0);
	VS3D_DrawAABBNodeBounds(eff, mesh->aabb_root);
	glPopAttrib();
}

void VS3D_DrawMeshBoundingVolume(RenderEffect3D *eff, M4Mesh *mesh)
{
	SFVec3f c, s;

	if (mesh->aabb_root) {
		VS3D_DrawMeshAABBTreeBounds(eff, mesh);
		return;
	}
	vec_diff(s, mesh->bounds.max_edge, mesh->bounds.min_edge);
	c.x = mesh->bounds.min_edge.x + s.x/2;
	c.y = mesh->bounds.min_edge.y + s.y/2;
	c.z = mesh->bounds.min_edge.z + s.z/2;

	/*polygon bit pushed for sphere bounds*/
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);
	glPushMatrix();
	
	glTranslatef(c.x, c.y, c.z);

	glDisable(GL_LIGHTING);
//	glColor3f(1.0, 1.0, 1.0);
	glColor3f(0, 0, 0);
	if (eff->surface->render->compositor->draw_bvol==M4_Bounds_Box) {
		glScalef(s.x, s.y, s.z);
		VS3D_DrawMeshIntern(eff, eff->surface->render->unit_bbox);
	} else {
		glScalef(mesh->bounds.radius, mesh->bounds.radius, mesh->bounds.radius);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		VS3D_DrawMeshIntern(eff, eff->surface->render->unit_bsphere);
	}
	glPopMatrix();
	glPopAttrib();
}

extern u32 time_spent_in_gl;

void VS3D_DrawMesh(RenderEffect3D *eff, M4Mesh *mesh, Bool do_normalize)
{
	if (do_normalize) glEnable(/*GL_NORMALIZE*/ GL_RESCALE_NORMAL);

	if (eff->surface->render->wiremode != M4_WireOnly) VS3D_DrawMeshIntern(eff, mesh);

	if (eff->surface->render->wiremode != M4_WireNone) {
		glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDisable(GL_LIGHTING);
		glColor3f(0.0, 0.0, 0.0);
		VS3D_DrawMeshIntern(eff, mesh);
		glPopAttrib();
	}

	if (eff->surface->render->compositor->draw_bvol) VS3D_DrawMeshBoundingVolume(eff, mesh);
}


/*only used for ILS/ILS2D or IFS2D outline*/
void VS3D_StrikeMesh(RenderEffect3D *eff, M4Mesh *mesh, Float width, u32 dash_style)
{
	u16 style;
	
	if (mesh->mesh_type != MESH_LINESET) return;

	switch (dash_style) {
	case M4StrikeDash:
		style = 0x1F1F;
		break;
	case M4StrikeDot:
		style = 0x3333;
		break;
	case M4StrikeDashDot:
		style = 0x6767;
		break;
	case M4StrikeDashDashDot:
		style = 0x33CF;
		break;
	case M4StrikeDashDotDot:
		style = 0x330F;
		break;
	default:
		style = 0;
		break;
	}

	width/=2;
	glLineWidth(width);
	if (style) {
		u32 factor = (u32) (width);
		if (!factor) factor = 1;
		glEnable(GL_LINE_STIPPLE);
		glLineStipple(factor, style); 
		VS3D_DrawMesh(eff, mesh, 0);
		glDisable (GL_LINE_STIPPLE);
	} else {
		VS3D_DrawMesh(eff, mesh, 0);
	}
}

void VS3D_SetMaterial2D(VisualSurface *surf, SFColor col, Float alpha)
{
	glDisable(GL_LIGHTING);
	if (alpha != 1.0) {
		glEnable(GL_BLEND);
		VS3D_SetAntiAlias(surf, 0);
	} else {
		glDisable(GL_BLEND);
		VS3D_SetAntiAlias(surf, surf->render->compositor->antiAlias ? 1 : 0);
	}
	glColor4f(col.red, col.green, col.blue, alpha);
}

void VS3D_SetAntiAlias(VisualSurface *surf, Bool bOn)
{
	if (bOn) {
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_POINT_SMOOTH);
		if (surf->render->poly_aa)
			glEnable(GL_POLYGON_SMOOTH);
		else
			glDisable(GL_POLYGON_SMOOTH);
	} else {
		glDisable(GL_LINE_SMOOTH);
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_POLYGON_SMOOTH);
	}
}

void VS3D_ClearSurface(VisualSurface *surf, SFColor color, Float alpha)
{
	glClearColor(color.red, color.green, color.blue, alpha);
	glClear(GL_COLOR_BUFFER_BIT);
}

void VS3D_DrawImage(VisualSurface *surf, Float pos_x, Float pos_y, u32 width, u32 height, u32 pixelformat, char *data, Float scale_x, Float scale_y)
{
	u32 gl_format;
	glPixelZoom(scale_x, scale_y);

	gl_format = 0;
	switch (pixelformat) {
	case M4PF_RGB_24:
		gl_format = GL_RGB;
		break;
	case M4PF_RGB_32:
	case M4PF_RGBA:
		gl_format = GL_RGBA;
		break;
	case M4PF_ARGB:
		if (!surf->render->compositor->hw_caps.bgra_texture) return;
		gl_format = GL_BGRA_EXT;
		break;
	default:
		return;
	}

	/*glRasterPos2f doesn't accept point outside the view volume (it invalidates all draw pixel, draw bitmap)
	so we move to the center of the local coord system, draw a NULL bitmap with raster pos displacement*/
	glRasterPos2f(0, 0);
	glBitmap(0, 0, 0, 0, pos_x, -pos_y, NULL);
	glDrawPixels(width, height, gl_format, GL_UNSIGNED_BYTE, data);
	glBitmap(0, 0, 0, 0, -pos_x, pos_y, NULL);

}

void VS3D_GetMatrix(VisualSurface *surf, u32 mat_type, Float *mat)
{
	switch (mat_type) {
	case MAT_MODELVIEW:
		glGetFloatv(GL_MODELVIEW_MATRIX, mat);
		break;
	case MAT_PROJECTION:
		glGetFloatv(GL_PROJECTION_MATRIX, mat);
		break;
	case MAT_TEXTURE:
		glGetFloatv(GL_TEXTURE_MATRIX, mat);
		break;
	}
}

void VS3D_SetMatrixMode(VisualSurface *surf, u32 mat_type)
{
	switch (mat_type) {
	case MAT_MODELVIEW:
		glMatrixMode(GL_MODELVIEW);
		break;
	case MAT_PROJECTION:
		glMatrixMode(GL_PROJECTION);
		break;
	case MAT_TEXTURE:
		glMatrixMode(GL_TEXTURE);
		break;
	}
}

void VS3D_ResetMatrix(VisualSurface *surf)
{
	glLoadIdentity();
}
void VS3D_PushMatrix(VisualSurface *surf)
{
	glPushMatrix();
}
void VS3D_MultMatrix(VisualSurface *surf, Float *mat)
{
	glMultMatrixf(mat);
}
void VS3D_PopMatrix(VisualSurface *surf)
{
	glPopMatrix();
}
void VS3D_LoadMatrix(VisualSurface *surf, Float *mat)
{
	glLoadMatrixf(mat);
}


void VS3D_SetClipper2D(VisualSurface *surf, M4Rect clip)
{
	Double g[4];
	g[2] = 0; 
	g[1] = 0; 
	g[3] = clip.x + clip.width; g[0] = -1; 
	glClipPlane(GL_CLIP_PLANE0, g); glEnable(GL_CLIP_PLANE0);
	g[3] = -clip.x; g[0] = 1;
	glClipPlane(GL_CLIP_PLANE1, g); glEnable(GL_CLIP_PLANE1);
	g[0] = 0;
	g[3] = clip.y; g[1] = -1; 
	glClipPlane(GL_CLIP_PLANE2, g); glEnable(GL_CLIP_PLANE2);
	g[3] = clip.height-clip.y; g[1] = 1; 
	glClipPlane(GL_CLIP_PLANE3, g); glEnable(GL_CLIP_PLANE3);
}

void VS3D_ResetClipper2D(VisualSurface *surf)
{
	glDisable(GL_CLIP_PLANE3);
	glDisable(GL_CLIP_PLANE2);
	glDisable(GL_CLIP_PLANE1);
	glDisable(GL_CLIP_PLANE0);
}

void VS3D_SetMaterial(VisualSurface *surf, u32 material_type, Float *rgba)
{
	switch (material_type) {
	case MATERIAL_AMBIENT:
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, rgba);
		break;
	case MATERIAL_DIFFUSE:
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, rgba);
		break;
	case MATERIAL_SPECULAR:
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, rgba);
		break;
	case MATERIAL_EMISSIVE:
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, rgba);
		break;
	case MATERIAL_NONE:
		glColor4fv(rgba);
		break;
	}
}

void VS3D_SetShininess(VisualSurface *surf, Float shininess)
{
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess * 128);
}

void VS3D_SetState(VisualSurface *surf, u32 flag_mask, Bool setOn)
{
	if (setOn) {
		if (flag_mask & F3D_LIGHT) glEnable(GL_LIGHTING);
		if (flag_mask & F3D_BLEND) glEnable(GL_BLEND);
		if (flag_mask & F3D_COLOR) glEnable(GL_COLOR_MATERIAL);
	} else {
		if (flag_mask & F3D_LIGHT) glDisable(GL_LIGHTING);
		if (flag_mask & F3D_BLEND) glDisable(GL_BLEND);
		if (flag_mask & F3D_COLOR) glDisable(GL_COLOR_MATERIAL | GL_COLOR_MATERIAL_FACE);
	}
}

Bool VS3D_AddSpotLight(VisualSurface *surf, Float ambientIntensity, SFVec3f attenuation, Float beamWidth, 
					   SFColor color, Float cutOffAngle, SFVec3f direction, Float intensity, SFVec3f location)
{
	Float vals[4];
	GLint iLight;

	if (!surf->num_lights) glEnable(GL_LIGHTING);
	if (surf->num_lights==surf->max_lights) return 0;
	iLight = GL_LIGHT0 + surf->num_lights;
	surf->num_lights++;
	glEnable(iLight);
	
	/*in case...*/
	vec_norm(direction);
	vals[0] = direction.x; vals[1] = direction.y; vals[2] = direction.z; vals[3] = 1;
	glLightfv(iLight, GL_SPOT_DIRECTION, vals);
	vals[0] = location.x; vals[1] = location.y; vals[2] = location.z; vals[3] = 1;
	glLightfv(iLight, GL_POSITION, vals);
	glLightf(iLight, GL_CONSTANT_ATTENUATION, attenuation.x ? attenuation.x : 1.0f);
	glLightf(iLight, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(iLight, GL_QUADRATIC_ATTENUATION, attenuation.z);
	vals[0] = color.red*intensity; vals[1] = color.green*intensity; vals[2] = color.blue*intensity; vals[3] = 1;
	glLightfv(iLight, GL_DIFFUSE, vals);
	glLightfv(iLight, GL_SPECULAR, vals);
	vals[0] = color.red*ambientIntensity; vals[1] = color.green*ambientIntensity; vals[2] = color.blue*ambientIntensity; vals[3] = 1;
	glLightfv(iLight, GL_AMBIENT, vals);

	glLightf(iLight, GL_SPOT_EXPONENT, (Float) (0.5 * log(0.5) / log(cos(beamWidth)) ) );
	glLightf(iLight, GL_SPOT_CUTOFF, 180*cutOffAngle/M_PI);
	return 1;
}

/*insert pointlight - returns 0 if too many lights*/
Bool VS3D_AddPointLight(VisualSurface *surf, Float ambientIntensity, SFVec3f attenuation, SFColor color, Float intensity, SFVec3f location)
{
	Float vals[4];
	u32 iLight;
	if (!surf->num_lights) glEnable(GL_LIGHTING);
	if (surf->num_lights==surf->max_lights) return 0;
	iLight = GL_LIGHT0 + surf->num_lights;
	surf->num_lights++;
	glEnable(iLight);

	vals[0] = location.x; vals[1] = location.y; vals[2] = location.z; vals[3] = 1;
	glLightfv(iLight, GL_POSITION, vals);

	glLightf(iLight, GL_CONSTANT_ATTENUATION, attenuation.x ? attenuation.x : 1.0f);
	glLightf(iLight, GL_LINEAR_ATTENUATION, attenuation.y);
	glLightf(iLight, GL_QUADRATIC_ATTENUATION, attenuation.z);
	vals[0] = color.red*intensity; vals[1] = color.green*intensity; vals[2] = color.blue*intensity; vals[3] = 1;
	glLightfv(iLight, GL_DIFFUSE, vals);
	glLightfv(iLight, GL_SPECULAR, vals);
	vals[0] = color.red*ambientIntensity; vals[1] = color.green*ambientIntensity; vals[2] = color.blue*ambientIntensity; vals[3] = 1;
	glLightfv(iLight, GL_AMBIENT, vals);

	glLightf(iLight, GL_SPOT_EXPONENT, 0);
	glLightf(iLight, GL_SPOT_CUTOFF, 180);
	return 1;
}

Bool VS3D_AddDirectionalLight(VisualSurface *surf, Float ambientIntensity, SFColor color, Float intensity, SFVec3f direction)
{
	Float vals[4];
	u32 iLight;
	if (!surf->num_lights) glEnable(GL_LIGHTING);
	if (surf->num_lights==surf->max_lights) return 0;
	iLight = GL_LIGHT0 + surf->num_lights;
	surf->num_lights++;
	glEnable(iLight);

	/*in case...*/
	vec_norm(direction);
	vals[0] = -direction.x; vals[1] = -direction.y; vals[2] = -direction.z; vals[3] = 0;
	glLightfv(iLight, GL_POSITION, vals);
	vals[0] = color.red*intensity; vals[1] = color.green*intensity; vals[2] = color.blue*intensity; vals[3] = 1;
	glLightfv(iLight, GL_DIFFUSE, vals);
	glLightfv(iLight, GL_SPECULAR, vals);
	vals[0] = color.red*ambientIntensity; vals[1] = color.green*ambientIntensity; vals[2] = color.blue*ambientIntensity; vals[3] = 1;
	glLightfv(iLight, GL_AMBIENT, vals);

	glLightf(iLight, GL_CONSTANT_ATTENUATION, 1.0f);
	glLightf(iLight, GL_LINEAR_ATTENUATION, 0);
	glLightf(iLight, GL_QUADRATIC_ATTENUATION, 0);
	glLightf(iLight, GL_SPOT_CUTOFF, 180);
	return 1;
}

void VS3D_RemoveLastLight(VisualSurface *surf)
{
	if (surf->num_lights) {
		glDisable(GL_LIGHT0+surf->num_lights-1);
		surf->num_lights--;
	}
}

void VS3D_ClearAllLights(VisualSurface *surf)
{
	u32 i;
	for (i=surf->num_lights; i>0; i--) {
		glDisable(GL_LIGHT0+i-1);
	}
	surf->num_lights = 0;
	glDisable(GL_LIGHTING);
}

void VS3D_SetFog(VisualSurface *surf, const char *type, SFColor color, Float density, Float visibility)
{
	Float vals[4];
	glEnable(GL_FOG);
	if (!type || !stricmp(type, "LINEAR")) glFogi(GL_FOG_MODE, GL_LINEAR);
	else if (!stricmp(type, "EXPONENTIAL")) glFogi(GL_FOG_MODE, GL_EXP);
	else if (!stricmp(type, "EXPONENTIAL2")) glFogi(GL_FOG_MODE, GL_EXP2);
	glFogf(GL_FOG_DENSITY, density);
	glFogf(GL_FOG_START, 0);
	glFogf(GL_FOG_END, visibility);
	vals[0] = color.red; vals[1] = color.green; vals[2] = color.blue; vals[0] = 1;
	glFogfv(GL_FOG_COLOR, vals);
	glHint(GL_FOG_HINT, surf->render->compositor->high_speed ? GL_FASTEST : GL_NICEST);
}

M4Err R3D_GetScreenBuffer(VisualRenderer *vr, M4VideoSurface *fb)
{
	u32 i, hy;
	char *tmp;
	Render3D *sr = (Render3D *)vr->user_priv;

	fb->video_buffer = malloc(sizeof(char)*3*sr->out_width * sr->out_height);
	fb->width = sr->out_width;
	fb->pitch = 3*sr->out_width;
	fb->height = sr->out_height;
	fb->os_handle = NULL;
	fb->pixel_format = M4PF_RGB_24;

	/*don't understand why I get BGR when using GL_RGB*/
	glReadPixels(sr->out_x, sr->out_y, sr->out_width, sr->out_height, GL_BGR_EXT, GL_UNSIGNED_BYTE, fb->video_buffer);

	/*flip image (openGL always handle image data bottom to top) */
	tmp = malloc(sizeof(char)*fb->pitch);
	hy = fb->height/2;
	for (i=0; i<hy; i++) {
		memcpy(tmp, fb->video_buffer+ i*fb->pitch, fb->pitch);
		memcpy(fb->video_buffer + i*fb->pitch, fb->video_buffer + (fb->height - 1 - i) * fb->pitch, fb->pitch);
		memcpy(fb->video_buffer + (fb->height - 1 - i) * fb->pitch, tmp, fb->pitch);
	}
	free(tmp);
	return M4OK;
}

M4Err R3D_ReleaseScreenBuffer(VisualRenderer *vr, M4VideoSurface *framebuffer)
{
	free(framebuffer->video_buffer);
	framebuffer->video_buffer = 0;
	return M4OK;
}
